// Copyright 2015, University of Freiburg,
// Chair of Algorithms and Data Structures.
// Author: Zhiwei Zhang <zhiwei.zhang@saturn.uni-freiburg.de>

#ifndef ZHIWEI_ZHANG_SOURCE_OTHER_ELIASDELTAENCODINGVECTOR_H_
#define ZHIWEI_ZHANG_SOURCE_OTHER_ELIASDELTAENCODINGVECTOR_H_

#include <vector>
#include "./EliasGammaEncodingVector.h"

using std::vector;

class EliasDeltaEncodingVector {
  // we can encode all the natural numbers (except 0) via elias delta encoding.

 public:

  // Encodes the gaps in @list via elias delta encoding.
  template<typename Numeric>
  static void encode(const vector<Numeric>& plaintext, Bitvector& encoded,
                       const bool& show = false) {
    Timer t;
    t.start();
    encoded.reserve(plaintext.size() * 10000);
    for (size_t i = 0; i < plaintext.size(); ++i) {
      /* since I use a gao list as input, so we don't need to calculate the gap.
       * if original list is given as input, just remove the comment mark and
       * add coment mark at line 41 - 54.
       *
       * Numeric gap = (i==0?plaintext[0]:plaintext[i]-plaintext[i-1]);
       * int prefix = floor(log2(gap)) + 1;
       * int numOfZerosOfPrefix = floor(log2(prefix));
       * for (int j = 0; j < numOfZerosOfPrefix; ++j) encoded.push_back(0);
       * for (int j = numOfZerosOfPrefix; j >= 0; --j) {
       *   encoded.push_back((prefix & (1 << j)) != 0);
       * }
       * for (int j = prefix - 2; j >= 0; --j) {
       *   encoded.push_back((gap & (1 << j)) != 0);
       *
       */
      int prefix = floor(log2(plaintext[i])) + 1;
      int numOfZerosOfPrefix = floor(log2(prefix));
      for (int j = 0; j < numOfZerosOfPrefix; ++j) encoded.push_back(0);
      for (int j = numOfZerosOfPrefix; j >= 0; --j) {
        // Get the value of the bit in r that is nofBits - j from the right.
        encoded.push_back((prefix & (1 << j)) != 0);
        // compress the binary number of prefix into result.
      }
      // the process before is EliasGammaEncoding, we should in the next compress the
      // the binary number of original number in result, then we finish the compress
      // of this number.
      for (int j = prefix - 2; j >= 0; --j) {
        // Get the value of the bit in r that is nofBits - j from the right.
        encoded.push_back((plaintext[i] & (1 << j)) != 0);
        // compress the binary number of gap into result.
      }
    }
    t.stop();
    if (show) {
      std::cout << plaintext.size() << "\t" << encoded.size()/8 << "\t"
      << t.getUSecs() << "\t";
    }
  }

  // Assumes @bits contains the gaps of a Doclist encoded with elias gamma,
  // Restores the Doclist.
  template<typename Numeric>
  static void decode(const Bitvector& bits, vector<Numeric>& decoded,
                         const bool& show = false) {
    Timer t;
    t.start();
    decoded.reserve(bits.size());
    size_t curPosition = 0;
    size_t nofZeros = 0;
    Numeric num = 0;
    while (curPosition < bits.size()) {
      while (bits[curPosition] == 0) ++curPosition;
      nofZeros = curPosition - nofZeros;
      // Read the number at the next bits.
      int prefix = -1;
      // from the first one to convert the binary number in decimal number.
      for (int j = nofZeros; j >= 0; --j) {
        if (bits[curPosition++] == 1) prefix += (1 << j);
      }
      /* since I use a gao list as input of compress, so we just need
       * to calculate the gap.
       * if original list is given as input of compress, just remove
       * the comment mark and add coment mark at line 93.
       *
       * num += (1 << prefix);
       *
       */
      num = (1 << prefix);
      for (int j = prefix - 1; j >= 0; --j) {
        if (bits[curPosition++] == 1) num += (1 << j);
      }
      // Add to the Doclist
      decoded.push_back(num);
      nofZeros = curPosition;
    }
    t.stop();
    if (show) std::cout << t.getUSecs() << std::endl;
  }
};

#endif  // ZHIWEI_ZHANG_SOURCE_OTHER_ELIASDELTAENCODINGVECTOR_H_
