// Copyright 2015, University of Freiburg,
// Chair of Algorithms and Data Structures.
// Author: Zhiwei Zhang <zhiwei.zhang@saturn.uni-freiburg.de>

#ifndef ZHIWEI_ZHANG_SOURCE_OTHER_ELIASGAMMAENCODINGVECTOR_H_
#define ZHIWEI_ZHANG_SOURCE_OTHER_ELIASGAMMAENCODINGVECTOR_H_

#include <math.h>
#include <iostream>
#include <vector>
#include "../Timer.h"

using std::vector;

typedef vector<bool> Bitvector;

class EliasGammaEncodingVector {
  // we can encode all the natural numbers (except 0) via elias gamma encoding.

 public:

  // Encodes the gaps in @list via elias gamma encoding.
  template<typename Numeric>
  static void encode(const vector<Numeric>& plaintext, Bitvector& encoded,
                       const bool& show = false) {
    Timer t;
    t.start();
    encoded.reserve(plaintext.size() * 10000);
    for (size_t i = 0; i < plaintext.size(); ++i) {
      /* since I use a gao list as input, so we don't need to calculate the gap.
       * if original list is given as input, just remove the comment mark and
       * add coment mark at line 43, then replace all the plaintext[i] with gap.
       *
       * Numeric gap = (i == 0 ? plaintext[0] : plaintext[i] - plaintext[i - 1]);
       * int numOfZeros = floor(log2(gap));
       * int numOfZeros = floor(log2(plaintext[i]));
       * for (size_t j = 0; j < numOfZeros; ++j) encoded.push_back(0);
       * for (int j = numOfZeros; j >= 0; --j) {
       *   encoded.push_back((gap & (1 << j)) != 0);
       *
       */
      int numOfZeros = floor(log2(plaintext[i]));
      for (size_t j = 0; j < numOfZeros; ++j) encoded.push_back(0);
      for (int j = numOfZeros; j >= 0; --j) {
        // That is the easy way to convert the decimal number in binary number.
        encoded.push_back((plaintext[i] & (1 << j)) != 0);
      }
    }
    t.stop();
    if (show) {
      std::cout << plaintext.size() << "\t" << encoded.size()/8 << "\t"
      << t.getUSecs() << "\t";
    }
  }

  // Assumes @bits contains the gaps of a Doclist encoded with elias gamma,
  // Restores the Doclist.
  template<typename Numeric>
  static void decode(const Bitvector& bits, vector<Numeric>& decoded,
                         const bool& show = false) {
    Timer t;
    t.start();
    decoded.reserve(bits.size());
    size_t curPosition = 0;
    size_t nofZeros = 0;
    /* since I use a gao list as input of compress, so we just need
     * to calculate the gap.
     * if original list is given as input of compress, just remove
     * the comment mark and add coment mark at line 75.
     *
     * Numeric num = 0;
     *
     */
    while (curPosition < bits.size()) {
      Numeric num = 0;
      while (bits[curPosition] == 0) ++curPosition;
      nofZeros = curPosition - nofZeros;
      // from the first one to convert the binary number in decimal number.
      for (int j = nofZeros; j >= 0; --j) {
        if (bits[curPosition++] == 1) num += (1 << j);
      }
      decoded.push_back(num);
      nofZeros = curPosition;
    }
    t.stop();
    if (show) std::cout << t.getUSecs() << std::endl;
  }
};

#endif  // ZHIWEI_ZHANG_SOURCE_OTHER_ELIASGAMMAENCODINGVECTOR_H_
